/***************************************************************************
 *             __________               __   ___.
 *   Open      \______   \ ____   ____ |  | _\_ |__   _______  ___
 *   Source     |       _//  _ \_/ ___\|  |/ /| __ \ /  _ \  \/  /
 *   Jukebox    |    |   (  <_> )  \___|    < | \_\ (  <_> > <  <
 *   Firmware   |____|_  /\____/ \___  >__|_ \|___  /\____/__/\_ \
 *                     \/            \/     \/    \/            \/
 *
 * Copyright (C) 2006 Katie Stafford
 *
 * All files in this archive are subject to the GNU General Public License.
 * See the file COPYING in the source tree root for full license agreement.
 *
 * This software is distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY
 * KIND, either express or implied.
 *
 ****************************************************************************/

#include "mixer.h"
#include "imath.h"
#include "datafile.h"

#define MAX_CHANNELS 20

enum {
    Q_MIX_FILL = 1,
    Q_MIX_STOP,
};

#if CONFIG_CODEC == SWCODEC && !defined(SIMULATOR)
#define pcm_play_stop RB pcm_play_stop
#define pcm_play_data RB pcm_play_data
#define pcm_is_playing RB pcm_is_playing
#else
#define pcm_play_stop() (void)0
#define pcm_play_data(a, b, c) (void)0
#define pcm_is_playing() (0)

#endif

#define NUM_FORMATS 10
static sound_format sound_formats[] = {
    { 11025, 1, 8, false, false, 0 },
    { 11025, 1, 16, true, false, 0 },
    { 11025, 2, 16, true, false, 0 },
    { 11025, 1, 16, true, true, 0 },
    { 11025, 2, 16, true, true, 0 },

    { 22050, 1, 8, false, false, 0 },
    { 22050, 1, 16, true, false, 0 },
    { 22050, 2, 16, true, false, 0 },
    { 22050, 1, 16, true, true, 0 },
    { 22050, 2, 16, true, true, 0 },
};

static uint8 *mixbuf[2];
static int mixpage;

static int num_channels IBSS_ATTR;

static mix_channel mix_channels[MAX_CHANNELS] IBSS_ATTR;

static int curstamp;

static struct event_queue mix_queue;
static long mix_stack[(DEFAULT_STACK_SIZE + 0x200)/sizeof(long)] IBSS_ATTR;
static struct thread_entry* mix_thread_ent;

static int mix_buffer_valid;
static int mixbuf_size = 512;

/* defined in asm block */
extern uint32 mix_internal_formats[];

void mix_thread(void) ICODE_ATTR;

static void post_fill() {
    RB queue_post(&mix_queue, Q_MIX_FILL, NULL);
}

void mix_init(void) {
    int i;

    for (i = 0; i < NUM_FORMATS; i++) {
        sound_formats[i].internal_format_desc = mix_internal_formats[i];
    }

    mixbuf[0] = malloc(mixbuf_size << 3);
    mixbuf[1] = malloc(mixbuf_size << 3);
    
    RB queue_init(&mix_queue, true);
    RB queue_post(&mix_queue, Q_MIX_FILL, NULL);
    
    
    mix_thread_ent = RB create_thread(mix_thread, mix_stack, sizeof(mix_stack), "rocken_mix"
                                      IF_PRIO(, PRIORITY_PLAYBACK));
    
    RB yield();
}

void mix_fini(void) {
    RB queue_post(&mix_queue, Q_MIX_STOP, NULL);
    while (mix_thread_ent) RB yield();

    RB queue_delete(&mix_queue);

    while (num_channels) {
        channel_halt(mix_channels[0].chan);
    }

    pcm_play_stop();
    while (pcm_is_playing())
        RB yield();

    free(mixbuf[0]);
    free(mixbuf[1]);
}

void mix_gametick(void) {
    mix_channel* mchan = &mix_channels[num_channels];
    while((mchan--) != mix_channels) {
        CALLMETH(mchan->chan, gametick);
    }
}

sound_format* mix_find_format(int freq, int channels, bool sign, bool big, int bitspersample) {
    int i;
    sound_format* cf = sound_formats;
    for (i = 0; i < NUM_FORMATS; i++, cf++) {
        if (freq == cf->freq && channels == cf->channels && 
            sign == cf->sign && bitspersample == cf->bitspersample &&
            (bitspersample == 8 || big == cf->big))
            return cf;
    }
    debugf("ifmt %d %d %d %d %d",freq, channels, bitspersample, sign, big);
    return NULL;
}

int load_wav(const char* fname, sound_buffer* sb) {

    int fd, fsize;
    int i;
    
    /* from the fmt chunk */
    int encoding = 0, channels = 0, freq = 0;
    int byterate = 0, blockalign = 0, bitspersample = 0;

    uint32 riffsig, flen, wavesig, ctype, clen;
    uint8 tdat[32];
    
    sound_format* sfmt = NULL;
    uint8* sdat = NULL;
    uint32 slen;

    buffer tbuf;


    tbuf.data = tdat;
    tbuf.pos = 0;
    fd = RB open(fname, O_RDONLY);
    if (fd < 0) {
        return -1;
    }
    fsize = RB filesize(fd);

    tbuf.len = RB read(fd, tdat, 12);

    riffsig = readu32(&tbuf);
    flen = readu32le(&tbuf);
    wavesig = readu32(&tbuf);

    if (riffsig != 0x52494646 /* 'RIFF' */ || wavesig != 0x57415645 /* 'WAVE' */) {
        RB close(fd);
        return -1;
    }

    while (1) {
        tbuf.pos = 0;
        tbuf.data = tdat;
        tbuf.len = RB read(fd, tdat, 8);
        if (tbuf.len < 8)
            break;
        ctype = readu32(&tbuf);
        clen = readu32le(&tbuf);
        if (clen == 0) clen = fsize - RB lseek(fd, 0, SEEK_CUR);
        if (clen <= 0) break;
        tbuf.data = malloc(clen);

        if (!tbuf.data) {
            if (sdat)
                free(sdat);
            RB close(fd);
            return -3;
        }
        tbuf.pos = 0;
        tbuf.len = RB read(fd, tbuf.data, clen);
        if (ctype == 0x666d7420) { /* 'fmt ' */
            if (tbuf.len >= 16) {
                encoding = readu16le(&tbuf);
                channels = readu16le(&tbuf);
                freq = readu32le(&tbuf);
                byterate = readu32le(&tbuf);
                blockalign = readu16le(&tbuf);
                bitspersample = readu16le(&tbuf);
                
                sfmt = mix_find_format(freq, channels, bitspersample != 8, false, bitspersample);
            }
            free(tbuf.data);
        } else if (ctype == 0x64617461) { /* 'data' */
            if (sdat)
                free(sdat);
            sdat = tbuf.data;
            slen = tbuf.len;
        } else {
            free(tbuf.data);
        }
    }
    
    RB close(fd);

    /* only support encoding 1 (pcm) */
    if (!sfmt || !sdat || encoding != 1) {
        if (sdat)
            free(sdat);
        return -1;
    }
    /*
#ifndef ROCKBOX_LITTLE_ENDIAN
    if (bitspersample == 16) {
        uint8* upos = sdat;
        uint8* end = (sdat + ((slen+3)&~3));
        while (upos < end) {
            uint8* tend = upos + 8192;
            if (tend > end) tend = end;
#ifdef CPU_COLDFIRE
            asm(
                "1:\n"
                "move.l  (%[pos]), %%d0\n"
                "move.l  %%d0, %%d1\n"
                "lsr.l   #8, %%d0\n"
                "lsl.l   #8, %%d1\n"
                "andi.l  #0x00ff00ff, %%d0\n"
                "andi.l  #0xff00ff00, %%d1\n"
                "or.l    %%d1, %%d0\n"
                "move.l  %%d0, (%[pos])+\n"
                "cmp.l   %[pos], %[end]\n"
                "blo.s   1b\n"
                : [pos] "+a" (upos)
                : [end] "d" (tend)
                : "d0", "d1");
#else
            while (upos < tend) {
                uint16 d = *(uint16*)upos;
                d = d>>8 | d<<8;
                *(uint16*)upos = d;
                upos += 2;
            }
#endif            
            RB yield();
        }
    }
#endif
    */
    sb->fmt = sfmt;
    sb->buf = sdat;
    sb->nsamps = slen / (sfmt->channels * (sfmt->bitspersample/8));
    return 0;
}

channel_group* channel_group_new(int size) {
    int tsize = sizeof(channel_group) + size*sizeof(channel*);
    channel_group* ret = (channel_group*)malloc(tsize);
    RB memset(ret, 0, tsize);
    ret->size = size;
    ret->num_active = 0;
    return ret;
}

void channel_init(void* o) {
    channel* self = (channel*)o;
    self->mixchan = NULL;
    self->group = NULL;
    self->stamp = 0;
}

void channel_fini(void* o) {
    channel* self = (channel*)o;
    channel_halt(self);
}

void channel_group_free(channel_group* g) {
    channel_group_halt(g);
    free(g);
}

void channel_group_halt(channel_group* g) {
    while (g->num_active) {
        channel_halt(g->channels[0]);
    }
}
channel* channel_group_oldest(channel_group* g) {
    int j;
    if (!g->num_active) 
        return NULL;

    channel* oldest = g->channels[0];
    for (j = 1; j < g->num_active; j++) {
        if (g->channels[j]->stamp < oldest->stamp)
            oldest = g->channels[j];
    }

    return oldest;
}

void channel_set_buffer(channel* self, sound_buffer* buf) {
    int samps = buf->nsamps;
    samps &= ~1; /* always have even number of samples */

    if (buf->fmt) {
        self->mixchan->pos = buf->buf;

        self->mixchan->end = self->mixchan->pos + 
            (buf->nsamps * (buf->fmt->channels * (buf->fmt->bitspersample/8)));
    } else {
        self->mixchan->pos = self->mixchan->end = NULL;
    }

}

int channel_start(channel* self, channel_group* g, sound_buffer* buf) {
    mix_channel* mchan;

    if (self->mixchan)
        channel_halt(self);

    if (g->num_active == g->size) {
        return CHANNEL_GROUP_FULL;
    }
    if (num_channels == MAX_CHANNELS || !buf->fmt)
        return CHANNEL_SYS_FULL;
    
    mchan = &mix_channels[num_channels++];

    mchan->internal_format_desc = buf->fmt->internal_format_desc;

    self->mixchan = mchan;
    mchan->chan = self;

    channel_set_buffer(self, buf);
    mchan->lrvol = 0x7fff7fff;

    self->stamp = ++curstamp;
    self->group = g;

    g->channels[g->num_active++] = self;

    object_ref(self);

    CALLMETH(self, gametick);

    post_fill();

    return CHANNEL_STARTED;
}

void channel_halt(channel* self) {
    mix_channel *mchan, *end;
    mchan = self->mixchan;

    if (mchan) {
        CALLMETH(self, halted);
        end = &mix_channels[--num_channels];
        if (end != mchan) {
            RB memcpy(mchan, end, sizeof(mix_channel));
            mchan->chan->mixchan = mchan;
        }
        self->mixchan = NULL;
        object_deref(self);
    }

    if (self->group) {
        int j;
        for (j = 0; j < self->group->num_active; j++) {
            if (self == self->group->channels[j]) {
                self->group->channels[j] = self->group->channels[--self->group->num_active];
                self->group = NULL;
                return;
            }
        }
    }
}

static int __attribute__((used)) _mix_channel_end(mix_channel* mchan)  {
    channel* chan = mchan->chan;
    
    CALLMETH(chan, buffer_end);


    if (mchan->pos == mchan->end) {
        channel_halt(chan);
        return 1;
    }
    return 0;
}


#ifdef CPU_COLDFIRE

extern void mix_sound(sint16* dst);

#define MIX_SAMPLE_A(LL, RR)                                                                    \
"    move.l   (%a0)+, %d1        | chan->lrvol -> d1 [LLRR]            \n" /* 2(1/0) */         \
"    mac.w    %"#LL", %d1u, %acc0  | acc0 += AL * L                     \n" /* 1(0/0) */        \
"    mac.w    %"#RR", %d1l, %acc1  | acc1 += AR * R                     \n" /* 1(0/0) */

#define MIX_SAMPLE_B(LL, RR)                                                                            \
"    mac.w    %"#LL", %d1u, %acc2  | acc2 += BL * L                    \n" /* 1(0/0) */                 \
"    mac.w    %"#RR", %d1l, %acc3  | acc3 += BR * R                    \n" /* 1(0/0) */                 \
"    move.l   %a1, -16(%a0)      | a1 -> chan->pos                     \n" /* 1(0/1) */                 \
"    addq.l   #4, %a0            | skip ptr to channel obj             \n" /* 1(0/0) */                 \
"    subq.l   #1, %d2                                                  \n" /* 1(0/0) */                 \
"    jbne     .mix_channel_loop                                        \n" /* 2(0/0) | 3(0/0) */        \
"    jbra     .write_sample                                            \n"                              \
"                                                                      \n"

asm (
".section .icode, \"ax\", @progbits                                    \n"
".type mix_sound, @function                                            \n"
"mix_sound:                                                            \n"
"    lea.l    -24(%sp), %sp                                            \n" /* 1(0/0) */
"    movem.l  %d2-%d5/%a2-%a3, (%sp)                                   \n" /* 7(0/6) */
"    move.l   28(%sp), %a2                                             \n" /* 2(1/0) */
"    move.l   mixbuf_size, %d3                                         \n" /* 2(1/0) */
"    lsr.l    #1, %d3                                                  \n" /* 1(0/0) */
"                                                                      \n"
"    ||  put the EMAC in fractional, saturating, 16-bit rounding mode  \n"
"    move.l   #0xe0, %macsr                                            \n" /* 1(0/0) */
"    move.l   num_channels, %d4                                        \n" /* 2(1/0) */
"                                                                      \n"
".mix_sample_loop:                                                     \n" 
"    move.l   %d4, %d2                                                 \n" /* 1(0/0) */
"    jbeq     .no_more_channels                                        \n" /* 3(0/0) | 1(0/0) */
"    lea.l    mix_channels, %a0                                        \n" /* 1(0/0) */
"                                                                      \n"
".mix_channel_loop:                                                    \n" 
"    move.l   (%a0)+, %a1        | chan->pos -> a1                     \n" /* 2(1/0) */
"    move.l   (%a0)+, %d0        | chan->end -> d0                     \n" /* 2(1/0) */
"    cmp.l    %a1, %d0           | done playing?                       \n" /* 1(0/0) */
"    jbeq     .channel_end                                             \n" /* 3(0/0) | 1(0/0) */
"    move.l   (%a0)+, %a3                                              \n" /* 2(1/0) */
"    jmp      (%a3)                                                    \n"
"                                                                      \n"

".mix_u8_mono_half:                                                    \n"
"    clr.l    %d0                | 0000                                \n" /* 1(0/0) */
"    move.b   (%a1)+, %d0        | 000A                                \n" /* 3(1/0) */
"    lsl.l    #8, %d0            | 00A0                                \n" /* 1(0/0) */
"    eori.l   #0x00008000, %d0   | convert to signed                   \n"
MIX_SAMPLE_A(d0l, d0l)
MIX_SAMPLE_B(d0l, d0l)

".mix_s16_mono_half_le:                                                \n"
"    clr.l    %d0                | 0000                                \n" /* 1(0/0) */
"    move.w   (%a1)+, %d0        | 00AA                                \n" /* 2(1/0) */
"    move.l   #0x00ff00ff, %d1                                         \n"
"    and.l    %d0, %d1                                                 \n"
"    eor.l    %d1, %d0                                                 \n"
"    lsl.l    #8,  %d1                                                 \n"
"    lsr.l    #8,  %d0                                                 \n"
"    or.l     %d1, %d0                                                 \n"
MIX_SAMPLE_A(d0l, d0l)
MIX_SAMPLE_B(d0l, d0l)

".mix_s16_stereo_half_le:                                              \n"
"    move.l   (%a1)+, %d0        | ALAR                                \n" /* 2(1/0) */
"    move.l   #0x00ff00ff, %d1                                         \n"
"    and.l    %d0, %d1                                                 \n"
"    eor.l    %d1, %d0                                                 \n"
"    lsl.l    #8,  %d1                                                 \n"
"    lsr.l    #8,  %d0                                                 \n"
"    or.l     %d1, %d0                                                 \n"
MIX_SAMPLE_A(d0u, d0l)
MIX_SAMPLE_B(d0u, d0l)

".mix_s16_mono_half_be:                                                \n"
"    clr.l    %d0                | 0000                                \n" /* 1(0/0) */
"    move.w   (%a1)+, %d0        | 00AA                                \n" /* 2(1/0) */
MIX_SAMPLE_A(d0l, d0l)
MIX_SAMPLE_B(d0l, d0l)

".mix_s16_stereo_half_be:                                              \n"
"    move.l   (%a1)+, %d0        | ALAR                                \n" /* 2(1/0) */
MIX_SAMPLE_A(d0u, d0l)
MIX_SAMPLE_B(d0u, d0l)

".mix_u8_mono_full:                                                    \n"
"    clr.l    %d0                | 0000                                \n" /* 1(0/0) */
"    move.b   (%a1)+, %d0        | 000A                                \n" /* 3(1/0) */
"    swap     %d0                | 0A00                                \n" /* 1(0/0) */
"    move.b   (%a1)+, %d0        | 0A0B                                \n" /* 3(1/0) */
"    lsl.l    #8, %d0            | A0B0                                \n" /* 1(0/0) */
"    eori.l   #0x80008000, %d0   | convert to signed                   \n"
MIX_SAMPLE_A(d0u, d0u)
MIX_SAMPLE_B(d0l, d0l)

".mix_s16_mono_full_le:                                                \n"
"    move.l   (%a1)+, %d0        | AABB                                \n" /* 2(1/0) */
"    move.l   #0x00ff00ff, %d1                                         \n"
"    and.l    %d0, %d1                                                 \n"
"    eor.l    %d1, %d0                                                 \n"
"    lsl.l    #8,  %d1                                                 \n"
"    lsr.l    #8,  %d0                                                 \n"
"    or.l     %d1, %d0                                                 \n"
MIX_SAMPLE_A(d0u, d0u)
MIX_SAMPLE_B(d0l, d0l)

".mix_s16_stereo_full_le:                                              \n"
"    move.l   (%a1)+, %d0        | AABB                                \n" /* 2(1/0) */
"    move.l   #0x00ff00ff, %d1                                         \n"
"    and.l    %d0, %d1                                                 \n"
"    eor.l    %d1, %d0                                                 \n"
"    lsl.l    #8,  %d1                                                 \n"
"    lsr.l    #8,  %d0                                                 \n"
"    or.l     %d1, %d0                                                 \n"
MIX_SAMPLE_A(d0u, d0l)
"    move.l   (%a1)+, %d0        | AABB                                \n" /* 2(1/0) */
"    move.l   #0x00ff00ff, %d1                                         \n"
"    and.l    %d0, %d1                                                 \n"
"    eor.l    %d1, %d0                                                 \n"
"    lsl.l    #8,  %d1                                                 \n"
"    lsr.l    #8,  %d0                                                 \n"
"    or.l     %d1, %d0                                                 \n"
MIX_SAMPLE_B(d0u, d0l)

".mix_s16_mono_full_be:                                                \n"
"    move.l   (%a1)+, %d0        | AABB                                \n" /* 2(1/0) */
MIX_SAMPLE_A(d0u, d0u)
MIX_SAMPLE_B(d0l, d0l)

".mix_s16_stereo_full_be:                                              \n"
"    move.l   (%a1)+, %d0        | AABB                                \n" /* 2(1/0) */
MIX_SAMPLE_A(d0u, d0l)
"    move.l   (%a1)+, %d0        | AABB                                \n" /* 2(1/0) */
MIX_SAMPLE_B(d0u, d0l)

"    move.l   (%a0)+, %d1        | chan->lrvol -> d1 [LLRR]            \n" /* 2(1/0) */
"    move.l   %a1, -16(%a0)      | a1 -> chan->pos                     \n" /* 1(0/1) */
"    mac.w    %d0u, %d1u, %acc0  | acc0 += A * L                       \n" /* 1(0/0) */
"    mac.w    %d0u, %d1l, %acc1  | acc1 += A * R                       \n" /* 1(0/0) */
"    mac.w    %d0l, %d1u, %acc2  | acc2 += B * L                       \n" /* 1(0/0) */
"    mac.w    %d0l, %d1l, %acc3  | acc3 += B * R                       \n" /* 1(0/0) */
"    addq.l   #4, %a0            | skip ptr to channel obj             \n" /* 1(0/0) */
"    subq.l   #1, %d2                                                  \n" /* 1(0/0) */
"    jbne     .mix_channel_loop                                        \n" /* 2(0/0) | 3(0/0) */
"                                                                      \n"
".write_sample:                                                        \n"
"    movclr.l %acc0, %d0         | 00LL                                \n" /* 2(0/0) */
"    movclr.l %acc1, %d1         | 00RR                                \n" /* 2(0/0) */
"                                                                      \n"
"    swap     %d0                | LL00                                \n" /* 1(0/0) */
"    or.l     %d1, %d0           | LLRR                                \n" /* 1(0/0) */
"    move.l   %d0, (%a2)+                                              \n" /* 1(0/1) */
"    move.l   %d0, (%a2)+                                              \n" /* 1(0/1) */
"                                                                      \n"
"    movclr.l %acc2, %d0         | 00LL                                \n" /* 2(0/0) */
"    movclr.l %acc3, %d1         | 00RR                                \n" /* 2(0/0) */
"                                                                      \n"
"    swap     %d0                | LL00                                \n" /* 1(0/0) */
"    or.l     %d1, %d0           | LLRR                                \n" /* 1(0/0) */
"    move.l   %d0, (%a2)+                                              \n" /* 1(0/1) */
"    move.l   %d0, (%a2)+                                              \n" /* 1(0/1) */
"                                                                      \n"
"    subq.l   #1, %d3                                                  \n" /* 1(0/0) */
"    jbne     .mix_sample_loop                                         \n" /* 2(0/0) | 3(0/0) */
"                                                                      \n"
".mix_done:                                                            \n"
"    movem.l  (%sp), %d2-%d5/%a2-%a3                                   \n" /* 7(0/6) */
"    lea.l    24(%sp), %sp                                             \n" /* 1(0/0) */
"    rts                                                               \n" /* 5(1/0) */
"                                                                      \n"
".no_more_channels:                                                    \n"
"    clr.l    (%a2)+                                                   \n" /* 1(0/1) */
"    clr.l    (%a2)+                                                   \n" /* 1(0/1) */
"    clr.l    (%a2)+                                                   \n" /* 1(0/1) */
"    clr.l    (%a2)+                                                   \n" /* 1(0/1) */
"    subq.l   #1, %d3                                                  \n" /* 1(0/0) */
"    jbne     .no_more_channels                                        \n" /* 2(0/0) | 3(0/0) */
"    jbra     .mix_done                                                \n" /* 2(0/0) */
"                                                                      \n"
".channel_end:                                                         \n"
"    lea.l    -8(%a0), %a0                                             \n" /* 1(0/0) */
"    lea.l    -24(%sp), %sp                                            \n" /* 1(0/0) */
"    move.l   %macsr, %d4                                              \n" /* 1(0/0) */
"    movem.l  %d2-%d4/%a0/%a2-%a3, (%sp)                               \n" /* 8(0/7) */
"    move.l   %a0, -(%sp)                                              \n" /* 1(0/1) */
"    jbsr     _mix_channel_end                                         \n" /* 3(0/1) */
"    movem.l  4(%sp), %d2-%d4/%a0/%a2-%a3                              \n" /* 8(7/0) */
"    move.l   %d4, %macsr                                              \n" /* 1(0/0) */
"    lea.l    28(%sp), %sp                                             \n" /* 1(0/0) */
"    move.l   num_channels, %d4                                        \n" /* 2(1/0) */
"    sub.l    %d0, %d2                                                 \n"
"    jbne     .mix_channel_loop                                        \n" /* 3(0/0) | 1(0/0) */
"    jbra     .write_sample                                            \n" /* 3(0/0) */
"                                                                      \n"
".section .rodata                                                      \n"
"mix_internal_formats:                                                 \n"
"    .long .mix_u8_mono_half                                           \n"
"    .long .mix_s16_mono_half_le                                       \n"
"    .long .mix_s16_stereo_half_le                                     \n"
"    .long .mix_s16_mono_half_be                                       \n"
"    .long .mix_s16_stereo_half_be                                     \n"
"    .long .mix_u8_mono_full                                           \n"
"    .long .mix_s16_mono_full_le                                       \n"
"    .long .mix_s16_stereo_full_le                                     \n"
"    .long .mix_s16_mono_full_be                                       \n"
"    .long .mix_s16_stereo_full_be                                     \n"
);
#else

/* C reference implementation
   Not very elegant, but relatively simple to understand.
*/

uint32 mix_internal_formats[] = { 0, 1, 2, 3, 4, 5 };
 
static void mix_sound(sint16* dst) {
    int i, j;
    for (i = mixbuf_size; i--;) {
        int salval, sarval; /* source sample a, left & right */
        int sblval, sbrval; /* source sample b, left & right */
        int alval = 0, arval = 0; /* dest sample a, left & right */
        int blval = 0, brval = 0; /* dest sample b, left & right */

        mix_channel* ch = mix_channels;
        for (j = num_channels; j--; ch++) {
            if (ch->pos == ch->end) {
                if (_mix_channel_end(ch)) {
                    --ch;
                    continue;
                }
            }
            switch (ch->internal_format_desc) {
            case 0: /* unsigned 8-bit mono, half rate */
                sblval = sbrval = salval = sarval = 0x8000 ^ ((*ch->pos++) << 8);
                break;
            case 1: /* signed 16-bit mono, half rate */
                sblval = sbrval = salval = sarval = (*((sint16*)ch->pos)++);
                break;
            case 2: /* signed 16-bit stereo, half rate */
                sblval = salval = (*((sint16*)ch->pos)++);
                sbrval = sarval = (*((sint16*)ch->pos)++);
                break;
            case 3: /* unsigned 8-bit mono, full rate */
                salval = sarval = 0x8000 ^ ((*ch->pos++) << 8);
                sblval = sbrval = 0x8000 ^ ((*ch->pos++) << 8);
                break;
            case 4: /* signed 16-bit mono, full rate */
                salval = sarval = (*((sint16*)ch->pos)++);
                sblval = sbrval = (*((sint16*)ch->pos)++);
                break;
            case 5: /* signed 16-bit stereo, full rate */
                salval = (*((sint16*)ch->pos)++);
                sarval = (*((sint16*)ch->pos)++);
                sblval = (*((sint16*)ch->pos)++);
                sbrval = (*((sint16*)ch->pos)++);
                break;

            }
            alval += (salval * (short)(ch->lrvol>>16))>>15;
            arval += (sarval * (short)(ch->lrvol))>>15;
            blval += (sblval * (short)(ch->lrvol>>16))>>15;
            brval += (sbrval * (short)(ch->lrvol))>>15;
        }
        
        if (alval > 32767)
            alval = 32767;
        else if (alval < -32768)
            alval = -32768;
        if (arval > 32767)
            arval = 32767;
        else if (arval < -32768)
            arval = -32768;
        *dst++ = alval;
        *dst++ = arval;
        *dst++ = alval;
        *dst++ = arval;

        if (blval > 32767)
            blval = 32767;
        else if (blval < -32768)
            blval = -32768;
        if (brval > 32767)
            brval = 32767;
        else if (brval < -32768)
            brval = -32768;
        *dst++ = blval;
        *dst++ = brval;
        *dst++ = blval;
        *dst++ = brval;
    }
}                       

#endif

void mix_get_more(uint8 ** start, size_t *size) {
    if (!mix_buffer_valid) {
	*size = 0;
	return;
    }
    *size = (mixbuf_size << 3);
    *start = mixbuf[mixpage];
    mixpage ^= 1;
    mix_buffer_valid = 0;
    post_fill();
}

void mix_thread(void) {
    struct event ev;
    while (1) {
        RB queue_wait_w_tmo(&mix_queue, &ev, 1000);

        switch (ev.id) {
        case Q_MIX_FILL:
            if (!mix_buffer_valid) {
                mix_channel* mchan = &mix_channels[num_channels];
                while((mchan--) != mix_channels) {
                    CALLMETH(mchan->chan, premix);
                }
                mix_sound((uint16 *) mixbuf[mixpage]);
                mix_buffer_valid = 1;
            }
            if (!pcm_is_playing()) {
                pcm_play_data(mix_get_more, NULL, 0);
            }
            break;

        case Q_MIX_STOP:
            /* signal that we have stopped */
            mix_thread_ent = NULL;
            RB remove_thread(NULL);
            /* this won't return */
            RB yield();
            break;

        case SYS_USB_CONNECTED:
            pcm_play_stop();
            RB usb_acknowledge(SYS_USB_CONNECTED_ACK);
            while(1) {
                RB queue_wait_w_tmo(&mix_queue, &ev, 10000);
                if(ev.id == SYS_USB_DISCONNECTED) {
                    RB usb_acknowledge(SYS_USB_DISCONNECTED_ACK);
                    break;
                }
            }
            break;
        }
    }
}


